#include "Domain.hpp"

#include <cmath>
#include <mpi.h>
#include "Vector.hpp"

Domain::Domain() 
    : p_pos(config.domainBufferSize)
    , p_vel(config.domainBufferSize)
    , p_frc(config.domainBufferSize)
{}

void Domain::insert(const Vector& pos, const Vector& vel) {
    p_pos.insert(pos);
    p_vel.insert(vel);
    p_frc.insert(Vector());
}

void intersect(const Vector& p1, const Vector& d1, const Vector& p2, const Vector& d2, double& k1, double& k2) {
    k1 = ((p2.x-p1.x) * d2.y - (p2.y-p1.y) * d2.x) / (d1.x * d2.y - d1.y * d2.x);
    k2 = ((p1.x-p2.x) * d1.y - (p1.y-p2.y) * d1.x) / (d2.x * d1.y - d2.y * d1.x);
}

/** reflects L on N */
Vector reflect(const Vector& L, const Vector& N) {
    const Vector& n = N.normalize();
    return n * (2 * (n.dot(L))) - L;
}

void Domain::tick(double dt) {
    // do not create threads for empty domains
    if(count() == 0) return;

    int x, y; Grid::getPosition(x, y);

    int neighbours[9] = {
        Grid::getRank(x - 1, y - 1),
        Grid::getRank(x - 1, y    ),
        Grid::getRank(x - 1, y + 1),
        Grid::getRank(x    , y - 1),
        Grid::getRank(x    , y    ),
        Grid::getRank(x    , y + 1),
        Grid::getRank(x + 1, y - 1),
        Grid::getRank(x + 1, y    ),
        Grid::getRank(x + 1, y + 1)
    };

    int pcount = count();

    // update particle forces
    #pragma omp parallel
    {
        #pragma omp for
        for(unsigned int i = 0; i < pcount; ++i) {
            p_frc[i].x = 0;
            p_frc[i].y = 0;
        }

        #pragma omp barrier

        #pragma omp for
        for(unsigned int i = 0; i < pcount; ++i) {

            // iterate neighbor domains
            for(int di = 0; di < 9; ++di) {
                if(neighbours[di] == -1) continue;
                Domain* d = sim->domains[neighbours[di]];

                // check every particle in the domain, except if we're in our own
                // domain, then only check those following the current particle (i)
                for(unsigned int j = (di == 4 ? i + 1 : 0); j < d->count(); ++j) {
                    Vector diff = p_pos[i] - d->p_pos[j];
                    double x = diff.len();
                    if(x > 0 && x < config.D + config.T) {
                        double force = 1 - (x - config.T) / config.D;
                        force = fmin(1, fmax(0, force));
                        force = config.F * pow(force, config.P);

                        const Vector& force_vector = diff.normalize() * force;
                        p_frc[i] += force_vector;
                        if(di == 4) {
                            #pragma omp critical
                            d->p_frc[j] -= force_vector;
                        }
                    }
                }
            }
        }

        #pragma omp barrier

        // update particle velocity and position, and collisions
        #pragma omp for
        for(unsigned int i = 0; i < pcount; ++i) {
            bool moved = false;

            Vector& pos = p_pos[i];
            Vector& vel = p_vel[i];
            Vector& frc = p_frc[i];

            vel += frc * dt;
            vel *= (1 - dt * 0.1);

            // check collision
            for(auto& mesh : sim->meshes) {
                for(int f = 0; f < mesh.fragments.count && !moved; ++f) {
                    MeshFragment& fragment = mesh.fragments[f];

                    double k_particle, k_fragment;
                    intersect(pos, vel * dt, fragment.pos, fragment.dir, k_particle, k_fragment);

                    // check intersection
                    if(k_particle >= 0 && k_particle <= 1 && k_fragment >= 0 && k_fragment <= 1) {
                        // Apply force to mesh
                        const Vector& f_ = vel.projectOnto(fragment.nor);
                        #pragma omp critical
                        mesh.force += f_;

                        // Reflect velocity
                        const Vector& new_vel = reflect(vel, fragment.dir) * (1-config.SurfaceDamping);

                        // Update particle position and velocity
                        pos = pos + vel * dt * k_particle + new_vel * dt * (1 - k_particle);
                        vel = new_vel;

                        // Mark as done
                        moved = true;

                        // printf("particle bounced at k_particle %.2f, k_fragment %.2f, fragment_id %d\n", k_particle, k_fragment, f);;
                    }

                }
                if(moved) break;
            }

            if(!moved) {
                pos += vel * dt;
            }

            // Border collisions, very simple
            if(pos.y >= 1) { pos.y = 2 - pos.y; vel.y *= -1; }
            if(pos.y <= 0) { pos.y =   - pos.y; vel.y *= -1; }

            // if(pos.x >= 1) { pos.x = 2 - pos.x; vel.x *= -1; }
            // if(pos.x <= 0) { pos.x =   - pos.x; vel.x *= -1; }

        }

    }

    // killing
    for(int i = pcount - 1; i >= 0; --i) {
        if(p_pos[i].x >= 1 || p_pos[i].x < 0) { 
            kill(i);
        }
    }
}

void Domain::kill(int i) {
    p_pos.remove(i);
    p_vel.remove(i);
    p_frc.remove(i);
}

void Domain::transfer() {
    for(int i = count() - 1; i >= 0; i--) {
        if(sim->grid.getDomain(p_pos[i]) != config.rank) {
            sim->insert(p_pos[i], p_vel[i]);
            p_pos.remove(i);
            p_vel.remove(i);
        }
    }
}

void Domain::write(int offsetOrTotalCount) {
    char filename[64];
    sprintf(filename, "%s/%06i", config.outdir, config.iteration);

    MPI_File fh;
    MPI_Status status;
    int i= MPI_File_open(MPI_COMM_WORLD, filename, MPI_MODE_CREATE | MPI_MODE_WRONLY, MPI_INFO_NULL, &fh);
    if(i) {
        printf("Error writing to %s (%i).\n", filename, i);
        return;
    }

    const int HEADER_LENGTH = sizeof(int) * (4 + config.clients) + sizeof(Vector);
    if(config.master) {
        int header[] = {
            HEADER_LENGTH,
            config.iteration,
            config.clients,
            offsetOrTotalCount
        };

        MPI_File_write(fh, header, 4, MPI_INT, &status);
        Vector& force = sim->meshes[0].force;
        MPI_File_write(fh, &force, 2, MPI_DOUBLE, &status);
    }

    MPI_File_seek(fh, sizeof(int) * 4 + sizeof(Vector) + sizeof(int) * config.rank, MPI_SEEK_SET);
    int c = count();
    MPI_File_write(fh, &c, 1, MPI_INT, &status);

    MPI_File_seek(fh, (config.master ? 0 : offsetOrTotalCount) * 4 * sizeof(double) + HEADER_LENGTH, MPI_SEEK_SET);
    MPI_File_write(fh, p_pos.ptr(), 2 * c, MPI_DOUBLE, NULL);
    MPI_File_write(fh, p_vel.ptr(), 2 * c, MPI_DOUBLE, NULL);

    MPI_File_close(&fh);
}

void Domain::sendTo(int remote) {
    MPI_Send(&p_pos.count, 1, MPI_INT, remote, TAG_COUNT, MPI_COMM_WORLD);
    // printf("[%02d] Sending pos count %d to %02d\n", config.rank, p_pos.count, remote);
    MPI_Send(p_pos.ptr(), 2 * p_pos.count, MPI_DOUBLE, remote, TAG_POSITIONS, MPI_COMM_WORLD);
    MPI_Send(p_vel.ptr(), 2 * p_vel.count, MPI_DOUBLE, remote, TAG_VELOCITIES, MPI_COMM_WORLD);
}

void Domain::broadcast(int root) {
    int c = count();
    MPI_Bcast(&c, 1, MPI_INT, root, MPI_COMM_WORLD);
    MPI_Bcast(p_pos.ptr(), 2 * c, MPI_DOUBLE, root, MPI_COMM_WORLD);
    MPI_Bcast(p_vel.ptr(), 2 * c, MPI_DOUBLE, root, MPI_COMM_WORLD);
    p_pos.count = c;
    p_vel.count = c;
}

void Domain::clear() {
    p_pos.clear();
    p_vel.clear();
    p_frc.clear();
}

unsigned int Domain::count() const {
    return p_pos.count;
}
