#include "Simulation.hpp"

#include <cstdlib>
#include <cmath>
#include <iostream>
#include <mpi.h>

#include "Particle.hpp"
#include "Domain.hpp"
#include "Config.hpp"
#include "Vector.hpp"

Simulation::Simulation()
    : receive_pos(config.domainBufferSize),
      receive_vel(config.domainBufferSize) {

    domains = new Domain*[config.clients];
    for(int i = 0; i < config.clients; ++i) {
        domains[i] = new Domain();
    }
}

void Simulation::init() {
}

void Simulation::tick(double dt) {
    // reset forces for meshes
    for(auto& mesh : meshes) {
        mesh.force = Vector();
    }

    // run spawners
    for(auto& spawner : spawners) {
        spawner.tick(dt);
    }

    // tick this process's domain
    domains[config.rank]->tick(dt);
}

void Simulation::sync() {
    if(config.clients == 1) return; // do not sync single process

    for(int i = 0; i < config.clients; ++i) {
        if(i != config.rank) {
            domains[i]->clear();
        }
    }

    int x, y;
    bool black, white;
    Grid::getPosition(x, y);

    // Step 1: Checkerboard
    black = ((x + y) % 2 == 0);
    white = !black;

    // Step 1.1: Black->White straight
    syncWith( 1,  0, black);
    syncWith(-1,  0, black);
    syncWith( 0, -1, black);
    syncWith( 0,  1, black);

    // Step 1.2: White->Black straight
    syncWith( 1,  0, white);
    syncWith(-1,  0, white);
    syncWith( 0, -1, white);
    syncWith( 0,  1, white);

    // Step 2: Striped
    black = (x % 2 == 0);
    white = !black;

    // Step 2.1: Black->White diagonal
    syncWith(-1,  1, black);
    syncWith(-1, -1, black);
    syncWith( 1, -1, black);
    syncWith( 1,  1, black);

    // Step 2.2: White->Black diagonal
    syncWith( 1,  1, white);
    syncWith(-1, -1, white);
    syncWith( 1, -1, white);
    syncWith(-1,  1, white);

    // move stuff to other domains
    domains[config.rank]->transfer();

    // sync mesh forces
    for(auto& mesh : sim->meshes) {
        Vector force;
        MPI_Reduce(&mesh.force, &force, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
        if(config.master & config.debug) {
            printf("Mesh force: %.2f | %.2f\n", force.x, force.y);
        }
        if(config.master) {
            mesh.force = force;
        }
    }
}

void Simulation::write() {
    int offset = 0;
    int count = domains[config.rank]->count();
    if(config.clients > 1) {
        // sync particle counts
        if(!config.master) {
            MPI_Recv(&offset, 1, MPI_INT, config.rank - 1, 0, MPI_COMM_WORLD, NULL);
            int nextOffset = offset + count;
            MPI_Send(&nextOffset, 1, MPI_INT, (config.rank + 1) % config.clients, 0, MPI_COMM_WORLD);
        } else {
            MPI_Send(&count, 1, MPI_INT, 1, 0, MPI_COMM_WORLD);
            MPI_Recv(&offset, 1, MPI_INT, config.clients - 1, 0, MPI_COMM_WORLD, NULL);

            // TODO: file header
            if(config.debug) {
                printf("TOTAL PARTICLE COUNT IN ITERATION %i: %i\n", config.iteration, offset);
            }
            config.totalCount = offset;
        }
    } else {
        config.totalCount = count;
    }

    // save stuff to file
    domains[config.rank]->write(offset);
}

void Simulation::syncAll(int from) {
    // each client broadcast to everyone
    for(int i = 0; i < config.clients; ++i) {
        domains[i]->broadcast(from == -1 ? i : from);
    }
}

void Simulation::syncWith(int dx, int dy, bool sending) {
    if(!sending) {
        dx = -dx; 
        dy = -dy;
    }

    int x, y, rx, ry;
    Grid::getPosition(x, y);
    rx = x + dx; 
    ry = y + dy;

    int rank = Grid::getRank(rx, ry);

    if(rank < 0 || rank >= config.clients) return; // do not send out of boundaries

    if(sending) {
        domains[config.rank]->sendTo(rank);
    } else {
        receiveFrom(rank);
    }
}

void Simulation::receiveFrom(int rank) {
    receive_pos.clear();
    receive_vel.clear();

    int count;
    MPI_Recv(&count, 1, MPI_INT, rank, TAG_COUNT, MPI_COMM_WORLD, NULL);
    // printf("[%02d] Received pos count %d from %02d\n", config.rank, count, rank);
    MPI_Recv(receive_pos.ptr(), 2 * count, MPI_DOUBLE, rank, TAG_POSITIONS, MPI_COMM_WORLD, NULL);
    MPI_Recv(receive_vel.ptr(), 2 * count, MPI_DOUBLE, rank, TAG_VELOCITIES, MPI_COMM_WORLD, NULL);

    for(int i = 0; i < count; ++i) {
        insert(receive_pos[i], receive_vel[i]);
    }
}

void Simulation::insert(const Vector& pos, const Vector& vel) {
    int domain = grid.getDomain(pos);
    domains[domain]->insert(pos, vel);
}

Simulation* sim = 0;

