#include <stdio.h>
#include <omp.h>
#include <mpi.h>
#include <list>
#include <cmath>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>

// #include <SFML/Window.hpp>
// #include <SFML/Graphics.hpp>

#include "Simulation.hpp"
#include "Particle.hpp"
#include "Vector.hpp"
#include "Domain.hpp"
#include "Utils.hpp"

void print_usage() {
    printf("Usage: fluidsim [options]\n"
        "Simulation parameters:\n"
        "   -i, --iterations  Maximum number of iterations\n"
        "   -t, --time        Maximum computation time, in seconds\n"
        "   -r, --rows        Number of grid rows\n"
        "   -b, --buffer      Particle buffer size per domain\n"
        "   -o, --out         Specify output directory (Default: ../data)\n"
        "   -s, --step        Time step (0.001)\n"
        "   -d, --damping     Surface Damping (0.8)\n"
        "   -S, --speed       Start speed (2)\n"
        "   -R, --rate        Emission rate, particles per iteration (100)\n"
        "   -M, --measure     Disable IO for performance measurement\n"
        "Model parameters:\n"
        "   -P, --power       Force power\n"
        "   -F, --force       Force factor\n"
        "   -D, --distance    Influence distance\n"
        "   -T, --threshold   Influence threshold\n"
        "Program control:\n"
        "   -h, --help        Display this help\n"
        "   -v, --verbose     Print debug information\n"
        "");
    exit(1);
}

int main(int argc, char** argv) {
    bool help = false;
    int max_iterations = 0;
    int max_time = 0;
    char* out = 0;
    int rows = 1;

    static struct option long_options[] =
    {
        {"help", no_argument, (int*)&help, 'h'},
        {"rows", required_argument, 0, 'r'},
        {"iterations", required_argument, 0, 'i'},
        {"time",  required_argument, 0, 't'},
        {"buffer", required_argument, 0, 'b'},
        {"measure", no_argument, (int*)&config.measure, 'M'},
        {"out", required_argument, 0, 'o'},
        {"verbose", no_argument, (int*)&config.debug, 'v'},
        {"power", required_argument, 0, 'P'},
        {"force", required_argument, 0, 'F'},
        {"distance", required_argument, 0, 'D'},
        {"threshold", required_argument, 0, 'T'},
        {"step", required_argument, 0, 's'},
        {"damping", required_argument, 0, 'd'},
        {"speed", required_argument, 0, 'S'},
        {"rate", required_argument, 0, 'R'},
        {0, 0, 0, 0}
    };

    while(true) {
        int option_index = 0;
        int c = getopt_long(argc, argv, "Mho:i:t:b:vr:P:F:D:T:d:S:R:s:", long_options, &option_index);

        if(c == -1) break;

        switch(c) {
            case 0:
            case '?':
                if(long_options[option_index].flag != 0) break;
                printf ("option %s", long_options[option_index].name);
                if (optarg)
                    printf (" with arg %s", optarg);
                printf ("\n");
                break;

            case 'o':
                config.outdir = optarg; break;
            case 'i':
                max_iterations = atoi(optarg); break;
            case 't':
                max_time = atoi(optarg); break;
            case 'b':
                config.domainBufferSize = atoi(optarg); break;
            case 'h':
                help = true; break;
            case 'M':
                config.measure = true; break;
            case 'v':
                config.debug = true; break;
            case 'r':
                rows = atoi(optarg); break;
            case 'P':
                config.P = atof(optarg);
                if(config.P <= 0) {
                    printf("Error: Force Power must be greater than 0.");
                    exit(1);
                }
                break;
            case 'F':
                config.F = atof(optarg); break;
            case 'D':
                config.D = atof(optarg);
                if(config.D <= 0) {
                    printf("Error: Influence Distance must be greater than 0.");
                    exit(1);
                }
                break;
            case 'T': 
                config.T = atof(optarg); break;
            case 'd': 
                config.SurfaceDamping = atof(optarg); break;
            case 'S': 
                config.StartSpeed = atof(optarg); break;
            case 'R': 
                config.EmissionRate = atof(optarg); break;
            case 's': 
                config.TimeStep = atof(optarg); break;
        }
    }

    if(help) {
        print_usage();
    }

    // Random seed
    srand((unsigned)time(NULL));

    // Initialize MPI
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &config.rank);
    MPI_Comm_size(MPI_COMM_WORLD, &config.clients);
    if(config.clients == 0) config.clients = 1;
    config.master = (config.rank == 0);

    if(config.clients % rows != 0) {
        printf("Error: invalid row count. Cannot split %d processes evenly into %d rows.\n", config.clients, rows);
        exit(1);
    }

    config.gridColumns = config.clients / rows;
    config.gridRows = rows;

    if(config.master) {
        // Create output directory
        char dir[128];
        sprintf(dir, "mkdir -p %s/; rm %s/* 2> /dev/null", config.outdir, config.outdir);
        system(dir);

        printf("===========================================\n");
        printf("  Running on %d nodes in a  %dx%d grid.\n", config.clients, config.gridColumns, config.gridRows);
        if(max_iterations) {
            printf("  Maximum iteration count: %d.\n", max_iterations);
        }
        if(max_time) {
            printf("  Maximum computation time: %d seconds.\n", max_time);
        }
        printf("===========================================\n");
        printf("  D=%.f  T=%.f  P=%.f  F=%.f  \n", config.D, config.T, config.P, config.F);
        printf("  s=%.f  d=%.f  S=%.f  R=%.f  b=%i  \n", config.TimeStep, config.SurfaceDamping, config.StartSpeed, config.EmissionRate, config.domainBufferSize);
        printf("===========================================\n");
    }

    char hostname[256];
    gethostname(hostname, sizeof(hostname));

    if(config.debug)
        printf("%02d: PID %d on %s\n", config.rank, getpid(), hostname);

    // Initialize Simulation
    sim = new Simulation();
    sim->init();

    // Create meshes
    Mesh m(7, 0.3);
    m.addPoint(Vector(0.2,  0.5));
    m.addPoint(Vector(0.25, 0.46));
    m.addPoint(Vector(0.3, 0.45));
    m.addPoint(Vector(0.4, 0.45));
    m.addPoint(Vector(0.6,  0.485));
    m.addPoint(Vector(0.9,  0.55));
    m.addPoint(Vector(0.5,  0.51));
    // for(int i = 0; i < 20; ++i) {
    //     m.addPoint(Vector(0.5 + 0.2 * cos(i/20.0*M_PI*2), 0.5 + 0.2 * sin(i/20.0*M_PI*2)));
    // }
    m.finish();
    sim->meshes.push_back(m);

    Spawner spawner(Area(Vector(0, 0), Vector(0, 1)), 
        config.EmissionRate / config.TimeStep, 
        Area(Vector(config.StartSpeed, -0.0), Vector(config.StartSpeed, 0.0)));

    sim->spawners.push_back(spawner);

    // Spawn initial particles in master
    // if(config.master) {
    //     for(int i = 0; i < 1000;) {
    //         Vector p(drand() * 0.25, drand());
    //         if(m.isOutside(p)) {
    //             sim->insert(p, Vector(10 * drand(), drand()));
    //             ++i;
    //         }
    //     }
    // }
                
    // for(float y = 0; y < 1; y += 0.01) {
    //     sim->insert(Vector(0.1, y), Vector(10, 0));
    // }

    // Propagate all particles to correct processes
    // sim->syncAll(0);

    // Iterate
    double total_start_time = MPI_Wtime();

    for(int i = 0;; ++i) {
        double start_time = MPI_Wtime();
        config.iteration = i;

        // create output directory
        // if(config.master) {
        //     char dir[64];
        //     sprintf(dir, "mkdir -p %s/%05i/", config.outdir, config.iteration);
        //     system(dir);
        // }
        // MPI_Barrier(MPI_COMM_WORLD);

        sim->tick(config.TimeStep);

        sim->sync();

        // Write simulation status
        if(!config.measure) {
            sim->write();

            char filename[64];
            sprintf(filename, "%s/status", config.outdir);
            MPI_File fh;
            MPI_Status status;
            MPI_File_open(MPI_COMM_WORLD, filename, MPI_MODE_CREATE | MPI_MODE_WRONLY, MPI_INFO_NULL, &fh);

            int data_len = 4;
            int* data = new int[data_len];
            data[0] = config.iteration;
            data[1] = config.gridColumns;
            data[2] = config.gridRows;
            data[3] = sim->meshes.size();

            if(config.master) {
                MPI_File_write(fh, data, data_len, MPI_INT, &status);
                for(auto& mesh : sim->meshes) {
                    int c = mesh.points.size();
                    MPI_File_write(fh, &c, 1, MPI_INT, &status);
                    MPI_File_write(fh, &mesh.points[0], c * 2, MPI_DOUBLE, &status);
                }
            }

            MPI_File_close(&fh);
        }

        double duration = MPI_Wtime() - start_time;
        double time = MPI_Wtime() - total_start_time;

        if(config.master && !config.debug) {
            printf("\r> Clients %i | Iteration %i | %i particles | Computation time %.1fs | Iteration time %.2fs ", config.clients, config.iteration, config.totalCount, time, duration);
            fflush(stdout); 
        }

        if(max_time > 0 && time > max_time) {
            if(config.master || config.debug) {
                printf("\nTime limit reached.\n");
            }
            break;
        }

        if(max_iterations > 0 && config.iteration >= max_iterations) {
            if(config.master || config.debug) {
                printf("\nIteration limit reached.\n");
            }
            break;
        }
    }

    MPI_Finalize();
}
