#include "Iteration.hpp"

#include <fstream>
#include "Modes.hpp"

Iteration::Iteration(int index)
    : m_index(index)
{
    load();
    generateVertexArray();
}

void Iteration::load() {
    char filename[64];
    sprintf(filename, "%s/%06i", outdir, m_index);

    std::ifstream file;
    file.open(filename, std::ios::binary);

    // read the file, display particles
    int headerLength=0, clients=0, iteration=0, particleCount=0;
    file.read((char*)&headerLength, sizeof(int));
    file.read((char*)&iteration, sizeof(int));
    file.read((char*)&clients, sizeof(int));
    file.read((char*)&particleCount, sizeof(int));
    file.read((char*)&mesh_x, sizeof(double) * 2);

    // printf("Header: %i, Clients: %i, Iteration: %i, Particles: %i\n", headerLength, clients, iteration, particleCount);
    // printf("Mesh force: %f %f\n", mesh_x, mesh_y);

    int* particleCounts = new int[clients];
    for(int i = 0; i < clients; ++i) {
        file.read((char*)&particleCounts[i], sizeof(int));
    }

    int jj = 0;
    for(unsigned int i = 0; i < clients; ++i) {
        double px, py, vx, vy;
        for(int j = 0; j < particleCounts[i]; ++j) {
            file.read((char*)&px, sizeof(double));
            file.read((char*)&py, sizeof(double));

            Particle* p = new Particle();
            p->pos = Vector(px, py);
            p->client = i;
            m_particles.push_back(p);
        }
        for(int j = 0; j < particleCounts[i]; ++j) {
            file.read((char*)&vx, sizeof(double));
            file.read((char*)&vy, sizeof(double));
            m_particles[jj]->vel = Vector(vx, vy);
            jj++;
        }
    }

    file.close();
}


double HueToRGB(double arg1, double arg2, double H) {
   if ( H < 0 ) H += 1;
   if ( H > 1 ) H -= 1;
   if ( ( 6 * H ) < 1 ) { return (arg1 + ( arg2 - arg1 ) * 6 * H); }
   if ( ( 2 * H ) < 1 ) { return arg2; }
   if ( ( 3 * H ) < 2 ) { return ( arg1 + ( arg2 - arg1 ) * ( ( 2.0 / 3.0 ) - H ) * 6 ); }
   return arg1;
}



const double D_EPSILON = 0.00000000000001; 
sf::Color fromHSL(double H, double S, double L) {
    float arg1, arg2;
    if(H!=1) H = fmod(H, 1);
    if(S!=1) S = fmod(S, 1);
    if(L!=1) L = fmod(L, 1);

    if (S <= D_EPSILON) {
        return sf::Color(L*255, L*255, L*255);
    } else {
        if ( L < 0.5 ) { arg2 = L * ( 1 + S ); }
        else { arg2 = ( L + S ) - ( S * L ); }
        arg1 = 2 * L - arg2;

        sf::Uint8 r =( 255 * HueToRGB( arg1, arg2, (H + 1.0/3.0 ) ) );
        sf::Uint8 g =( 255 * HueToRGB( arg1, arg2, H ) );
        sf::Uint8 b =( 255 * HueToRGB( arg1, arg2, (H - 1.0/3.0 ) ) );
        return sf::Color(r,g,b);
    }

}

void Iteration::generateVertexArray() {
    m_vertexArray.clear();
    
    if(ColorMode::current == ColorMode::VELOCITY) {
        m_vertexArray.setPrimitiveType(sf::Lines);
    } else if(ColorMode::current == ColorMode::PRETTY) {
        m_vertexArray.setPrimitiveType(sf::Quads);
    } else {
        m_vertexArray.setPrimitiveType(sf::Points);
    }

    sf::Color process_colors[] = {
        sf::Color::Red,
        sf::Color::Yellow,
        sf::Color::Cyan,
        sf::Color::Magenta,
        sf::Color(255, 128, 0),
        sf::Color::Green,
        sf::Color::Blue,
        sf::Color::White
    };

    for(auto p : m_particles) {
        sf::Color color = sf::Color::White;
        if(ColorMode::current == ColorMode::PROCESS) {
            color = process_colors[p->client % 7];
        } else if(ColorMode::current == ColorMode::VELOCITY) {
            color = fromHSL(p->vel.angle() / 2.0 / M_PI, 1, 0.5);
        }

        if(ColorMode::current == ColorMode::PRETTY) {
            double d = 0.03;
            int r = 50;
            sf::Color c(255, 255, 255, 10);
            m_vertexArray.append(sf::Vertex(sf::Vector2f(p->pos.x - d, p->pos.y - d), c, sf::Vector2f(0, 0)));
            m_vertexArray.append(sf::Vertex(sf::Vector2f(p->pos.x + d, p->pos.y - d), c, sf::Vector2f(r, 0)));
            m_vertexArray.append(sf::Vertex(sf::Vector2f(p->pos.x + d, p->pos.y + d), c, sf::Vector2f(r, r)));
            m_vertexArray.append(sf::Vertex(sf::Vector2f(p->pos.x - d, p->pos.y + d), c, sf::Vector2f(0, r)));
        } else {
            m_vertexArray.append(sf::Vertex(sf::Vector2f(p->pos.x, p->pos.y), color));

            if(ColorMode::current == ColorMode::VELOCITY) {
                m_vertexArray.append(sf::Vertex(sf::Vector2f(p->pos.x+p->vel.x/400.f, p->pos.y+p->vel.y/400.f), color));
            }
        }
    }
}

void Iteration::draw(sf::RenderTarget& target) {
    generateVertexArray();

    if(ColorMode::current == ColorMode::PRETTY) {
        sf::RenderStates states(sf::BlendAdd, sf::Transform::Identity, &blur, 0);
        target.draw(m_vertexArray, states);
    } else {
        target.draw(m_vertexArray);
    }

    if(GuiMode::current == GuiMode::GUI || GuiMode::current == GuiMode::GUIGRID) {
        sf::VertexArray arrow(sf::Lines);
        arrow.append(sf::Vertex(sf::Vector2f(0.5, 0.5), sf::Color::Green));
        arrow.append(sf::Vertex(sf::Vector2f(0.5 + mesh_x * 0.01, 0.5 + mesh_y * 0.01), sf::Color::Yellow));
        target.draw(arrow);
    }
}

int Iteration::getParticleCount() const {
    return m_particles.size();
}
