#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>

#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>

#include "Iteration.hpp"
#include "Vector.hpp"
#include "Modes.hpp"

/** Globals **/

int total_iteration_count = 0;
int grid_rows = 1, grid_colums = 1;
std::map<int, Iteration> iterations;
Iteration* current_iteration;
int current_iteration_index = 0;
sf::VertexArray mesh(sf::LinesStrip);

/** Helper **/

void readStatus() {
    std::ifstream file;
    char filename[64];
    sprintf(filename, "%s/status", outdir);
    file.open(filename, std::ios::binary);
    file.read((char*)&total_iteration_count, sizeof(int));
    file.read((char*)&grid_colums, sizeof(int));
    file.read((char*)&grid_rows, sizeof(int));

    sf::Color meshColor(255, 255, 255, 100);
    int meshes;
    file.read((char*)&meshes, sizeof(int));
    for(int m = 0; m < meshes; ++m) {
        int count;
        file.read((char*)&count, sizeof(int));
        for(int i = 0; i < count; ++i) {
            double dx, dy;
            file.read((char*)&dx, sizeof(double));
            file.read((char*)&dy, sizeof(double));
            mesh.append(sf::Vertex(sf::Vector2f(dx, dy), meshColor));
        }
    }

    file.close();
}

void drawGrid(sf::RenderTarget& target) {
    sf::Color gridColor(255, 255, 255, 100);
    sf::VertexArray grid(sf::Lines);
    for(float x = 0; x <= 1; x += (1.f / grid_colums)) {
        grid.append(sf::Vertex(sf::Vector2f(x, 0), gridColor));
        grid.append(sf::Vertex(sf::Vector2f(x, 1), gridColor));
    }
    for(float y = 0; y <= 1; y += (1.f / grid_rows)) {
        grid.append(sf::Vertex(sf::Vector2f(0, y), gridColor));
        grid.append(sf::Vertex(sf::Vector2f(1, y), gridColor));
    }

    if(GuiMode::current == GuiMode::GRID || GuiMode::current == GuiMode::GUIGRID) {
        target.draw(grid);
    }
}

/** Iterations **/

void setIteration(int i) {
    while(i < 0) i += total_iteration_count;
    while(i >= total_iteration_count) i -= total_iteration_count;

    if(iterations.count(i) == 0) {
        iterations.insert(std::pair<int,Iteration>(i,Iteration(i)));
    }
    current_iteration = &iterations[i];
    current_iteration_index = i;
}

/** Main function and loop **/

int main(int argc, char** argv) {
    if(argc >= 2) {
        outdir = argv[1];
    } else {
        outdir = "../data";
    }


    sf::RenderWindow window(sf::VideoMode(800, 800, 32), "FluidVis");
    sf::Font font;
    font.loadFromFile("data/DejaVuSansMono.ttf");

    blur.loadFromFile("data/blur.png");

    float totalTime = 0;
    float frameTime = 0;
    int fps = 20;

    sf::Clock clock;

    sf::View worldView(sf::FloatRect(-0.1, -0.1, 1.2, 1.2));
    const sf::View& defaultView = window.getDefaultView();

    readStatus();
    setIteration(0);

    while(window.isOpen()) {
        readStatus();
        float dt = clock.restart().asSeconds();
        totalTime += dt;
        frameTime += dt;

        if(State::current == State::RUNNING) {
            if(frameTime > 1.0 / fps) {
                frameTime = 0;
                setIteration(current_iteration_index + 1); 
            }
        } else if(State::current == State::LAST) {
            setIteration(total_iteration_count - 1);
        }

        sf::Event e;
        while(window.pollEvent(e)) {
            if(e.type == sf::Event::Closed) {
                window.close();
            } else if(e.type == sf::Event::KeyPressed) {
                if(e.key.code == sf::Keyboard::Left) {
                    setIteration(current_iteration_index - 1);
                } else if(e.key.code == sf::Keyboard::Right) {
                    setIteration(current_iteration_index + 1);
                } else if(e.key.code == sf::Keyboard::C) {
                    toggleMode(ColorMode::current);
                } else if(e.key.code == sf::Keyboard::G) {
                    toggleMode(GuiMode::current);
                } else if(e.key.code == sf::Keyboard::Space) {
                    if(e.key.shift) {
                        if(State::current == State::LAST) {
                            State::current = State::RUNNING;
                        } else {
                            State::current = State::LAST;
                        }
                    } else {
                        if(State::current == State::RUNNING) {
                            State::current = State::MANUAL;
                        } else {
                            State::current = State::RUNNING;
                        }
                    }
                } else if(e.key.code == sf::Keyboard::Add) {
                    fps = fps + 1;
                } else if(e.key.code == sf::Keyboard::Subtract) {
                    fps = (fps > 0 ? fps - 1 : 0);
                } else if(e.key.code == sf::Keyboard::BackSpace) {
                    setIteration(0);
                }
            }
        }


        window.clear();

        window.setView(worldView);
        current_iteration->draw(window);

        drawGrid(window);
        window.draw(mesh);
        if(GuiMode::current == GuiMode::GUI || GuiMode::current == GuiMode::GUIGRID) {

            window.setView(defaultView);
            sf::Text status(
                " Iteration: " + std::to_string(current_iteration_index) + "/" + std::to_string(total_iteration_count) + "\n"
                "     State: " + (State::current == State::RUNNING ? "running (" + std::to_string(fps) + " FPS)" : (State::current == State::MANUAL ? "paused" : "last")) + "\n"
                " Particles: " + std::to_string(current_iteration->getParticleCount()) + "\n"
                , font, 12);
            status.setPosition(10, 10);
            window.draw(status);

            sf::Text help(
                "[G]           Toggle UI/Grid\n"
                "[C]           Toggle Colors\n"
                "[Space]       Pause/Resume\n"
                "[Shift+Space] Snap to End\n"
                "[Left/Right]  Step Iteration\n"
                "[Backspace]   Rewind"
                , font, 12);
            help.setPosition(10, window.getSize().y - 10 - help.getLocalBounds().height);
            help.setColor(sf::Color(255, 255, 255, 128));
            window.draw(help);

            sf::RectangleShape rectangle;
            rectangle.setSize(sf::Vector2f(400, 10));
            rectangle.setOutlineColor(sf::Color::White);
            rectangle.setFillColor(sf::Color::Transparent);
            rectangle.setOutlineThickness(1);
            rectangle.setPosition(window.getSize().x - 420, window.getSize().y - 30);
            window.draw(rectangle);

            sf::RectangleShape rectangle2;
            rectangle2.setSize(sf::Vector2f(2, 10));
            rectangle2.setPosition(window.getSize().x - 420 + 400 * current_iteration_index / (total_iteration_count-1), window.getSize().y - 30);
            window.draw(rectangle2);
        }

        window.display();
    }

    return 0;
}